#include <nemo/str.h>
#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/getln.h>
#include <nemo/exit.h>
#include <nemo/open.h>
#include <nemo/unix.h>
#include <nemo/error.h>

#include "mess822.h"
#include "die.h"

const char USAGE[] = "file mimetype [ file mimetype ]\n";

static stralloc line = STRALLOC;
static unsigned int match;

static unsigned int flag_content_type = 0;
static sa_vector content_type_list = SA_VECTOR;

static unsigned int flag_content_encoding = 0;
static sa_vector content_encoding_list = SA_VECTOR;

static unsigned int flag_mime_version = 0;

static mess822_header h = MESS822_HEADER;
static mess822_action a[] = {
/* name                         flag                    c  value                   a  w  */
 { "content-type",              &flag_content_type,     0, &content_type_list,     0, 0 },
 { "mime-version",              &flag_mime_version,     0, 0,                      0, 0 },
 { "content-transfer-encoding", &flag_content_encoding, 0, &content_encoding_list, 0, 0 },
 { 0, 0, 0, 0, 0, 0 }
};

static stralloc body_content_type = STRALLOC;
static stralloc body_content_encoding = STRALLOC;

static stralloc content_type = STRALLOC;
static sa_vector content_type_keys = SA_VECTOR;
static sa_vector content_type_data = SA_VECTOR;

static stralloc boundary = STRALLOC;
static stralloc boundary_start = STRALLOC;
static stralloc boundary_end = STRALLOC;

static unsigned int boundary_exists_in_message = 0;

static void scan_for_boundary(void)
{
  unsigned int i;
  for (i = 0; i < content_type_keys.len; i++) {
    if (stralloc_equals(&content_type_keys.va[i], "boundary")) {  /* always look for last entry */
      if (!stralloc_copy(&boundary, &content_type_data.va[i])) die_nomem();
    }
  }
}

static void process_existing_content_type(void)
{
/* only interested in last instance of the content type header */
  if (!mess822_content_field(&content_type, &content_type_keys, &content_type_data, &content_type_list.va[content_type_list.len - 1])) die_nomem();
  if (stralloc_starts(&content_type, "multipart/")) {
    scan_for_boundary();
  }
}

#define BLOCK_SIZE 57

static void encode_binary_file(const char *filename)
{
  static stralloc unencoded = STRALLOC;
  static stralloc encoded = STRALLOC;
  static char file_buffer[1024];
  static char block[BLOCK_SIZE];

  djbio io_file;
  int fd;
  int r;

  fd = open_read(filename);
  if (fd < 0) die_open(filename);
  djbio_initread(&io_file, read, fd, file_buffer, sizeof(file_buffer));
  if (!stralloc_ready(&unencoded, BLOCK_SIZE)) die_nomem();
  if (!stralloc_erase(&unencoded)) die_nomem();
  for (;;) {
    r = djbio_get(&io_file, block, BLOCK_SIZE - unencoded.len);
    if (r < 0) die_read(filename);
    if (!r) break;  /* end of file */
    if (!stralloc_catb(&unencoded, block, (unsigned int)r)) die_nomem();
    if (unencoded.len < BLOCK_SIZE) continue;
    if (!stralloc_base64_encode(&encoded, &unencoded)) die_nomem();
    djbio_putsa(djbiofd_out, &encoded);
    djbio_puteol(djbiofd_out);
    if (!stralloc_erase(&unencoded)) die_nomem();
  }
  close(fd);
  if (unencoded.len) {
    if (!stralloc_base64_encode(&encoded, &unencoded)) die_nomem();
    djbio_putsa(djbiofd_out, &encoded);
    djbio_puteol(djbiofd_out);
  }
  djbio_puteol(djbiofd_out);
}

static void add_attachment(const char *pathname, const char *mime_type)
{
  static stralloc filename = STRALLOC;

  unsigned int pathlen;
  unsigned int start;

  pathlen = str_len(pathname);
  start = str_rchr(pathname, '/');
  if (start == pathlen) {
    start = 0;
  }
  else {
    start++;  /* skip over sep */
  }
  if (!stralloc_copys(&filename, &pathname[start])) die_nomem();
  stralloc_replace_non_printable(&filename, '_');
  stralloc_replace(&filename, '"', '_');
  djbio_putsa(djbiofd_out, &boundary_start);
  djbio_puteol(djbiofd_out);
  djbio_puts(djbiofd_out, "Content-Type: ");
  djbio_puts(djbiofd_out, mime_type);
  djbio_puteol(djbiofd_out);
  djbio_puts(djbiofd_out, "Content-Transfer-Encoding: Base64\n");
  djbio_puts(djbiofd_out, "Content-Disposition: attachment; filename=\"");
  djbio_putsa(djbiofd_out, &filename);
  djbio_puts(djbiofd_out, "\"\n");
  djbio_puteol(djbiofd_out);

  encode_binary_file(pathname);
}

int main(int argc, char **argv)
{
  int i;

  PROGRAM = argv[0];
  if ((argc < 3) || ((argc & 1) == 0)) die_usage();  /* must have attachement matched with mime type */

  if (!stralloc_erase(&boundary)) return 0;
  if (!sa_vector_erase(&content_type_list)) return 0;
/*
  process headers
*/
  if (!mess822_begin(&h, a)) die_nomem();
  for (;;) {
    if (getln(djbiofd_in, &line, &match, '\n') < 0) {
      if (errno == error_nomem) die_nomem();
      die_read("input");
    }
    if (!mess822_ok(&line)) break;
    if (!mess822_line(&h, &line)) die_nomem();
    if (flag_content_type) {  /* do not output content headers */
      flag_content_type = 0;
      if (!match) break;
      continue;
    }
    if (flag_content_encoding) {  /* do not output content transfer encoding */
      flag_content_encoding = 0;
      if (!match) break;
      continue;
    }
    djbio_putsa(djbiofd_out, &line);
    if (!match) break;
  }
  if (!mess822_end(&h)) die_nomem();
/*
  process mime headers
*/
  if (!flag_mime_version) {
    djbio_puts(djbiofd_out, "MIME-Version: 1.0 (822attach)\n");
    /* djbio_puts(djbiofd_out, "MIME-Version: 1.0\n"); */
  }
/*
  set up body content type, extract existing boundary
*/
  if (content_type_list.len) {
    process_existing_content_type();
    if (!stralloc_copy(&body_content_type, &content_type)) die_nomem();
  }
  else {
    if (!stralloc_copys(&body_content_type, "text/plain; charset=\"us-ascii\"")) die_nomem();
  }
  if (boundary.len) {
    boundary_exists_in_message = 1;
  }
  if (!boundary_exists_in_message) {
    if (!stralloc_copys(&content_type, "multipart/mixed")) die_nomem();
    if (!mess822_mime_delimiter(&boundary, "822attach-part")) die_nomem();
  }
/*
  set up content transfer encoding
*/
  if (content_encoding_list.len) {  /* grab last known value */
    if (!stralloc_copy(&body_content_encoding, &content_encoding_list.va[content_encoding_list.len - 1])) die_nomem();
  }
  else {  /* default value */
    if (!stralloc_copys(&body_content_encoding, "8bit")) die_nomem();
  }
/*
  output new headers
*/
  djbio_puts(djbiofd_out, "Content-Type: ");
  djbio_putsa(djbiofd_out, &content_type);
  djbio_puts(djbiofd_out, ";\n  boundary=\"");
  djbio_putsa(djbiofd_out, &boundary);
  djbio_puts(djbiofd_out, "\"\n\n");

  if (!stralloc_copys(&boundary_start, "--")) die_nomem();
  if (!stralloc_cat(&boundary_start, &boundary)) die_nomem();

  if (!stralloc_copys(&boundary_end, "--")) die_nomem();
  if (!stralloc_cat(&boundary_end, &boundary)) die_nomem();
  if (!stralloc_cats(&boundary_end, "--")) die_nomem();
/*
  process body
*/
  if (boundary_exists_in_message) {
    for (;;) {
      if (getln(djbiofd_in, &line, &match, '\n') < 0) {
        if (errno == error_nomem) die_nomem();
        die_read("input");
      }
      if (stralloc_start(&line, &boundary_end)) break;
      djbio_putsa(djbiofd_out, &line);
      if (!match) break;
    }
  }
  else {
    djbio_putsa(djbiofd_out, &boundary_start);
    djbio_puteol(djbiofd_out);
    djbio_puts(djbiofd_out, "Content-Type: ");
    djbio_putsa(djbiofd_out, &body_content_type);
    djbio_puteol(djbiofd_out);
    djbio_puts(djbiofd_out, "Content-Transfer-Encoding: ");
    djbio_putsa(djbiofd_out, &body_content_encoding);
    djbio_puteol(djbiofd_out);
    djbio_puteol(djbiofd_out);
    for (;;) {
      if (getln(djbiofd_in, &line, &match, '\n') < 0) die_read("input");
      djbio_putsa(djbiofd_out, &line);
      if (!match) break;
    }
  }
/*
  add attachments
*/
  for (i = 1; i < argc; i += 2) {
    add_attachment(argv[i], argv[i + 1]);
  }
/*
  finalise
*/
  djbio_putsa(djbiofd_out, &boundary_end);
  djbio_puteol(djbiofd_out);
  djbio_flush(djbiofd_out);
  _exit(0);
}
