#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/byte.h>
#include <nemo/str.h>
#include <nemo/strerr.h>
#include <nemo/djbiofd.h>
#include <nemo/getln.h>
#include <nemo/exit.h>
#include <nemo/tai.h>
#include <nemo/caldate.h>
#include <nemo/caltime.h>
#include <nemo/leapsecs.h>
#include <nemo/char.h>
#include <nemo/error.h>
#include <nemo/macro_unused.h>

#include "mess822.h"
#include "die.h"

#define LINE_LEN 78

static const char FATAL[] = "822print: fatal: ";

static void put_spaces(unsigned int len)
{
  while (len-- > 0) {
    djbio_put(djbiofd_out, " ", 1);
  }
}

static void put_format(const char *buf, unsigned int len)
{
  while (len > 1) {
    switch (*buf++) {
      case '_':
        djbio_put(djbiofd_out, "_\b", 2);
        break;
      case '+':
        djbio_put(djbiofd_out, buf, 1);
        djbio_put(djbiofd_out, "\b", 1);
        break;
      default:
        break;
    }
    djbio_put(djbiofd_out, buf++, 1);
    len -= 2;
  }
}

static void format(stralloc *out, const stralloc *in, const char *pre, char style1, char style2)
{
  unsigned int i;
  char ch;
  char ch2;

  if (!stralloc_copys(out, pre)) die_nomem();

  for (i = 0; i < in->len; i++) {
    ch = in->s[i];
    if (ch == '\n') ch = ' ';
    if (ch == '\t') ch = ' ';
    if ((ch >= 32) && (ch <= 126)) {
      if (!stralloc_append(out, &style1)) die_nomem();
      if (!stralloc_append(out, &ch)) die_nomem();
      continue;
    }

    if (!stralloc_append(out, &style2)) die_nomem();
    if (!stralloc_append(out, "\\")) die_nomem();
    ch2 = char_hex_chars[7 & (ch >> 6)];
    if (!stralloc_append(out, &style2)) die_nomem();
    if (!stralloc_append(out, &ch2)) die_nomem();
    ch2 = char_hex_chars[7 & (ch >> 3)];
    if (!stralloc_append(out, &style2)) die_nomem();
    if (!stralloc_append(out, &ch2)) die_nomem();
    ch2 = char_hex_chars[7 & ch];
    if (!stralloc_append(out, &style2)) die_nomem();
    if (!stralloc_append(out, &ch2)) die_nomem();
  }
}

static stralloc left_line = STRALLOC;
static stralloc right_line = STRALLOC;
static stralloc tmp = STRALLOC;
static stralloc blank = STRALLOC;

static void do_left(void)
{
  char *x;
  unsigned int xlen;
  unsigned int i;

  xlen = left_line.len / 2;
  x = left_line.s;
  while (xlen > LINE_LEN) {
    i = byte_rchr(x, LINE_LEN * 2, ' ') / 2;
    put_format(x, i * 2);
    djbio_put(djbiofd_out, "\n", 1);
    x += i * 2;
    xlen -= i;
  }
  put_format(x, xlen * 2);
  djbio_put(djbiofd_out, "\n", 1);
}

static void do_left_right(void)
{
  unsigned int left_xlen;
  unsigned int left_xpos;
  unsigned int right_xlen;
  unsigned int right_xpos;
  unsigned int i;
  unsigned int j;

  right_xlen = right_line.len / 2;
  left_xlen = left_line.len / 2;

  left_xpos = 0;
  right_xpos = 0;

  while (left_xlen + right_xlen > LINE_LEN) {
    i = 50;
    if (i > left_xlen) i = left_xlen;
    j = 50;
    if (j > right_xlen) j = right_xlen;
    if (i + j > 76) j = 76 - i;

    put_format(left_line.s + left_xpos, i * 2);
    put_spaces(78 - i - j);
    put_format(right_line.s + right_xpos, j * 2);
    djbio_put(djbiofd_out, "\n", 1);
    left_xlen -= i;
    left_xpos += 2 * i;
    right_xlen -= j;
    right_xpos += 2 * j;
  }

  put_format(left_line.s + left_xpos, left_xlen * 2);
  put_spaces(78 - left_xlen - right_xlen);
  put_format(right_line.s + right_xpos, right_xlen * 2);
  djbio_put(djbiofd_out, "\n", 1);
}

static void put_value(const stralloc *a, const char *pre)
{
  format(&right_line, a, "", '_', '=');
  format(&left_line, &blank, pre, '=', '_');
  do_left_right();
}

static void put_fields(const sa_vector *a)
{
  unsigned int j;

  for (j = 0; j < a->len; ++j) {
    format(&left_line, &a->va[j], "", '=', '_');
    do_left();
  }
}

static mess822_addrlist addrlist = MESS822_ADDRLIST;

static void put_addr(const sa_vector *a, const char *pre, char style)
{
  mess822_addrdata *addrdata;
  unsigned int i;
  unsigned int j;

  for (i = 0; i < a->len; i++) {
    if (!mess822_addrlist_scan(&addrlist, &a->va[i])) die_nomem();
    if (!stralloc_erase(&left_line)) die_nomem();
    if (!stralloc_erase(&right_line)) die_nomem();
    for (j = 0; j < addrlist.len; j++) {
      addrdata = &addrlist.ad[j];
      /* XXX: replace addr with nickname? */
      if (addrdata->type == '@') {
	format(&right_line, &addrdata->data, "", style, '_');
      }
      else if (addrdata->type == '(' || addrdata->type == ':') {
	format(&left_line, &addrdata->data, pre, '_', '=');
      }
    }
    if (!left_line.len) {
      format(&left_line, &blank, pre, '_', '=');
    }
    do_left_right();
  }
}

static char buf[FMT_CALTIME];

static void put_date(const stralloc *text, mess822_time *t)
{
  struct tai sec;
  struct caltime local;
  unsigned int blen;

  format(&left_line, text, "=D=a=t=e=:= ", '=', '=');
  if (!t->known) {
    do_left();
    return;
  }

  caltime_tai(&t->ct, &sec);
  caltime_local(&local, &sec, 0, 0);

  blen = caltime_fmt(&local, buf) - 6;
  if (!stralloc_copyb(&tmp, buf, blen)) die_nomem();
  format(&right_line, &tmp, "", '_', '=');
  do_left_right();
}

static void put_divider(const char *s)
{
  if (!stralloc_copys(&tmp, s)) die_nomem();
  format(&left_line, &tmp, "", '+', '+');
  do_left();
}

static sa_vector envelope = SA_VECTOR;
static sa_vector threading = SA_VECTOR;

static sa_vector mid = SA_VECTOR;
static sa_vector refs = SA_VECTOR;
static sa_vector returnpath = SA_VECTOR;

static sa_vector subject = SA_VECTOR;

static sa_vector dates = SA_VECTOR;
static mess822_time date;

static sa_vector to = SA_VECTOR;
static sa_vector cc = SA_VECTOR;
static sa_vector bcc = SA_VECTOR;
static sa_vector nrudt = SA_VECTOR;

static sa_vector from = SA_VECTOR;
static sa_vector sender = SA_VECTOR;
static sa_vector replyto = SA_VECTOR;
static sa_vector mailreplyto = SA_VECTOR;
static sa_vector followupto = SA_VECTOR;

static sa_vector misc = SA_VECTOR;

static mess822_action a[] = {
/*  name                                 f  copy        value         a  when  */
  { "date",                              0, 0,          &dates,       0, &date },
  { "subject",                           0, 0,          &subject,     0, 0 },
  { "to",                                0, 0,          &to,          0, 0 },
  { "cc",                                0, 0,          &cc,          0, 0 },
  { "bcc",                               0, 0,          &bcc,         0, 0 },
  { "apparently-to",                     0, 0,          &bcc,         0, 0 },
  { "notice-requested-upon-delivery-to", 0, 0,          &nrudt,       0, 0 },
  { "from",                              0, 0,          &from,        0, 0 },
  { "sender",                            0, 0,          &sender,      0, 0 },
  { "reply-to",                          0, 0,          &replyto,     0, 0 },
  { "mail-reply-to",                     0, 0,          &mailreplyto, 0, 0 },
  { "mail-followup-to",                  0, 0,          &followupto,  0, 0 },
  { "message-id",                        0, 0,          &mid,         0, 0 },
  { "references",                        0, 0,          &refs,        0, 0 },
  { "in-reply-to",                       0, &threading, 0,            0, 0 },
  { "return-path",                       0, 0,          &returnpath,  0, 0 },
  { "received",                          0, &envelope,  0,            0, 0 },
  { "delivered-to",                      0, &envelope,  0,            0, 0 },
  { "errors-to",                         0, &envelope,  0,            0, 0 },
  { "resent-sender",                     0, &envelope,  0,            0, 0 },
  { "resent-from",                       0, &envelope,  0,            0, 0 },
  { "resent-reply-to",                   0, &envelope,  0,            0, 0 },
  { "resent-to",                         0, &envelope,  0,            0, 0 },
  { "resent-cc",                         0, &envelope,  0,            0, 0 },
  { "resent-bcc",                        0, &envelope,  0,            0, 0 },
  { "resent-date",                       0, &envelope,  0,            0, 0 },
  { "resent-message-id",                 0, &envelope,  0,            0, 0 },
  { 0,                                   0, &misc,      0,            0, 0 }
};

static void output_headers(void)
{
  unsigned int i;

  for (i = 0; i < dates.len; i++) {
    put_date(&dates.va[i], &date);
  }

  for (i = 0; i < subject.len; i++) {
    put_value(&subject.va[i], "=S=u=b=j=e=c=t=:");
  }

  put_addr(&from, "=F=r=o=m=:= ", '+');

  put_addr(&to, "=T=o=:= ", '+');
  put_addr(&cc, "=C=c=:= ", '+');
  put_addr(&bcc, "=B=c=c=:= ", '+');

  put_divider("------------------------------------------------------------------------------");

  put_addr(&nrudt, "=N=o=t=i=c=e=-=R=e=q=u=e=s=t=e=d=-=U=p=o=n=-=D=e=l=i=v=e=r=y=-=T=o=:= ", '+');
  put_addr(&returnpath, "=R=e=t=u=r=n=-=P=a=t=h=:= ", '+');
  put_fields(&envelope);
  put_fields(&threading);
  put_addr(&refs, "=R=e=f=:= ", '_');
  put_addr(&mid, "=M=e=s=s=a=g=e=-=I=D=:= ", '_');

  put_divider("------------------------------------------------------------------------------");

  put_addr(&sender, "=S=e=n=d=e=r=:= ", '+');
  put_addr(&replyto, "=R=e=p=l=y=-=T=o=:= ", '+');
  put_addr(&mailreplyto, "=M=a=i=l=-=R=e=p=l=y=-=T=o=:= ", '+');
  put_addr(&followupto, "=M=a=i=l=-=F=o=l=l=o=w=u=p=-=T=o=:= ", '+');

  put_fields(&misc);

  put_divider("-- Message body --------------------------------------------------------------");
}

static stralloc line = STRALLOC;
static unsigned int match;

int main(int arc __UNUSED__, char **argv)
{
  PROGRAM = argv[0];
  if (leapsecs_init() < 0) die_leapsecs();

  if (!mess822_slurp_headers(djbiofd_in, &line, a)) {
    if (errno == error_proto) die_proto();
    if (errno == error_nomem) die_nomem();
    die_read("input");
  }
  output_headers();

  for (;;) {
    if (getln(djbiofd_in, &line, &match, '\n') < 0) {
      djbio_flush(djbiofd_out);
      if (errno == error_nomem) die_nomem();
      die_read("input");
    }
    djbio_putsa(djbiofd_out, &line);
    if (!match) break;
  }
  djbio_flush(djbiofd_out);
  _exit(0);
}
